# coding: utf-8

#
module Gomasio
  DEFAULT_BOARD_SIZE = 19

  #
  class Board
    attr_reader :size
    attr_accessor :order

    #
    @directions = []
    Offset = Struct.new(:down, :right) do
      def reverse
        Offset.new(-down, -right)
      end
    end

    #
    [-1, 0, 1].repeated_permutation(2) do |right, down|
      next if right == 0 && down == 0
      @directions << Offset.new(right, down)
    end

    #
    def initialize(board_size = DEFAULT_BOARD_SIZE)
      @size  = board_size
      @range = 0..(@size - 1)
      clear
    end

    #
    def clone
      instance = Board.new(@size)
      instance.load(@board.clone)
      instance
    end

    #
    def clear
      @board = Array.new(@size**2, Gomasio::Disk.empty)
    end

    #
    def count
      black = white = 0
      (@size**2).times do |index|
        next if @board[index].empty?
        @board[index].black? ? black += 1 : white += 1
      end
      [black, white]
    end

    #
    def full?
      (@size**2).times do |index|
        return false if @board[index].empty?
      end
      true
    end

    #
    def [](row, col)
      row, col = cleansing(row, col)
      fail OutOfRangeError.new unless include?(row, col)
      @board[encode(row, col)]
    end

    #
    def mount(row: nil, col: nil, disk: nil)
      row, col = cleansing(row, col)
      fail OutOfRangeError.new unless include?(row, col)
      fail InSamePlaceError.new unless self[row, col].empty?

      @board[encode(row, col)] = disk

      point = Struct.new(:row, :col)
      directions = Board.instance_variable_get(:@directions)
      location = directions.each_with_object(Hash.new(0)) do |diff, item|
        (1..4).each do |index|
          dict = point.new(row + diff.down * index, col + diff.right * index)
          break unless include?(dict.row, dict.col)
          break if self[dict.row, dict.col] != disk
          item[diff] = index
        end
      end
      [location, win?(location)]
    end

    #
    def include?(row, col)
      row, col = cleansing(row, col)
      @range.include?(row) && @range.include?(col)
    end

    #
    def to_matrix
      Array.new(@size).map{[]}.each_with_index do |item, row|
        @size.times do |col|
          item << @board[encode(row,col)]
        end
      end
    end

    protected

    #
    def load(board)
      @board = board if board.kind_of?(Array)
    end

    private

    #
    def cleansing(row, col)
      [row.to_i, col.to_i]
    end

    #
    def win?(location)
      [[1, 0], [0, 1], [1, 1], [1, -1]].each do |diff|
        offset = Offset.new(diff[0], diff[1])
        return true if (location[offset] + location[offset.reverse] >= 4)
      end
      false
    end

    #
    def encode(row, col)
      row * @size + col
    end

    #
    def decode(index)
      index.divmod(@size)
    end
  end
end
